/* escape.cc
 */
#include "osl/category/escape.h"
#include "osl/category/probTable.tcc"
#include "osl/move_action/store.h"
#include "osl/move_generator/escape.h"
#include "osl/pieceEval.h"
#include "osl/moveLogProbVector.h"
#include <string>

namespace osl
{
  namespace category
  {
    DEF_CATEGORY_TABLE(PtypePtypeProbTable, ImmediateEscape);
    template <Ptype PTYPE> 
    DEF_CATEGORY_TABLE(PtypeValueTable, Escape<PTYPE>);
  }
}

struct osl::category::EscapeCommon::Generator
{
  const CategoryEnv::effect_state_t *state;
  const PtypeValueTable *table;
  int limit;
  MoveLogProbVector *out;
  int *indices;
  Generator(const CategoryEnv::effect_state_t *s, const PtypeValueTable *t,
	    int l, MoveLogProbVector *o, int *io)
    : state(s), table(t), limit(l), out(o), indices(io)
  {
  }
  /** @param p ƨ */
  void operator()(Piece p)
  {
    const Player myself = state->getTurn();
    assert(getOwner(p) == myself);
    if (state->hasEffectBy(alt(myself),getPosition(p)))
    {
      MoveVector moves;
      move_action::Store store(moves);
      const Player opponent = alt(myself);
      const Piece attacker = 
	EffectUtil::cheapestPiece(opponent, *state, getPosition(p));
      if(myself==BLACK)
	move_generator::Escape<BLACK,CategoryEnv::effect_state_t,move_action::Store>::
	  generateMoves(*state, p, attacker, store);
      else
	move_generator::Escape<WHITE,CategoryEnv::effect_state_t,move_action::Store>::
	  generateMoves(*state, p, attacker, store);
      for (size_t i=0; i<moves.size(); ++i)
      {
	const int diff = PieceEval::
	  computeDiffAfterMoveForRP(*state, moves[i]);
	const bool supported 
	  = state->hasEffectBy(myself, getPosition(p));
	const size_t index = 
	  table->getIndex(supported, getPtype(attacker), diff);
	const int prob = (*table)[index];
	if (prob <= limit)
	{
	  out->push_back(moves[i], prob);
	  if (indices)
	    *indices++ = index;
	}
      }
    }
  }
};

template struct osl::category::Escape<osl::ROOK>;
template struct osl::category::Escape<osl::BISHOP>;
template struct osl::category::Escape<osl::GOLD>;
template struct osl::category::Escape<osl::SILVER>;
template struct osl::category::Escape<osl::KNIGHT>;
template struct osl::category::Escape<osl::LANCE>;
template struct osl::category::Escape<osl::PAWN>;


template <osl::Ptype PTYPE>
void osl::category::Escape<PTYPE>::
generate(const CategoryEnv& env, MoveLogProbVector& out, int *indices)
{
  EscapeCommon::Generator gen(env.state, &probTable, env.limit, &out, indices);
  if (env.state->getTurn() == BLACK)
    (*env.state).template forEachOnBoard<BLACK,PTYPE,EscapeCommon::Generator>
      (gen);
  else
    (*env.state).template forEachOnBoard<WHITE,PTYPE,EscapeCommon::Generator>
      (gen);
}

template <osl::Ptype PTYPE>
void osl::category::Escape<PTYPE>::
generate(const CategoryEnv& env, MoveLogProbVector& out)
{
  generate(env, out, 0);
}

template <osl::Ptype PTYPE>
const char * osl::category::Escape<PTYPE>::getName()
{
  static std::string name = std::string("Escape-")
    + PtypeTraits<PTYPE>::name();
  return name.c_str(); 
}


struct osl::category::ImmediateEscape::PositionAction
{
  const CategoryEnv::effect_state_t *state;
  Position attackFrom;
  int limit;
  MoveLogProbVector *out;
  int *indexOut;
  PositionAction(const CategoryEnv::effect_state_t *s, Position from,
		 int l, MoveLogProbVector *o, int *iout)
    : state(s), attackFrom(from), limit(l), out(o), indexOut(iout)
  {
  }
  template<Player P>
  void doAction(Piece attacker, Position pos)
  {
    assert(P==alt(state->getTurn()));
    const Piece p = state->getPieceAt(pos);	   // פγ⤷ʤ
    if (isPiece(p) && (getOwner(p) == state->getTurn()))
    {
      MoveVector moves;
      move_action::Store store(moves);
      move_generator::Escape<PlayerTraits<P>::opponent,CategoryEnv::effect_state_t,move_action::Store>::
	generateMoves(*state, p,state->getPieceAt(attackFrom), store);
      for (size_t i=0; i<moves.size(); ++i)
      {
	const int diff = PieceEval::
	  computeDiffAfterMoveForRP(*state, moves[i]);
	const bool supported 
	  = state->hasEffectBy(state->getTurn(), attackFrom);
	const size_t index = 
	  probTable.getIndex(getPtype(p), supported, getPtype(attacker), diff);
	const int prob = probTable[index];
	if (prob <= limit)
	{
	  out->push_back(moves[i], prob);
	  if (indexOut)		// FIXME: ٤ɤޤ
	    *indexOut++ = index;
	}
      }
    }
  }
};

void osl::category::ImmediateEscape::
generate(const CategoryEnv& env, MoveLogProbVector& out, int *indices)
{
  if (! env.history->hasLastMove()) 
    return;
  const Move lastMove=env.history->lastMove();
  const Position pos = getTo(lastMove);
  const Piece attacker = env.state->getPieceOnBoard(pos);
  assert(isPiece(attacker));
  assert(getOwner(attacker) == alt(env.state->getTurn()));

  PositionAction a(env.state, pos, env.limit, &out, indices);
  env.state->forEachEffectOfPiece(attacker,a);
}

void osl::category::ImmediateEscape::
generate(const CategoryEnv& env, MoveLogProbVector& out)
{
  generate(env, out, 0);
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
