/* tableAnalyzer.cc
 */
#include "osl/search/tableAnalyzer.h"
#include "osl/search/simpleHashTable.h"
#include "osl/search/simpleHashRecord.h"
#include "osl/move_generator/allMoves.h"
#include "osl/move_action/store.h"
#include "osl/moveVector.h"
#include "osl/applyMove.h"
#include "osl/evalCompareLarger.h"
#include "osl/moveLogProb.h"
#include "osl/doUndoMoveStack.h"
#include <iostream>
#include <iomanip>
#include <map>
#include <set>
#include <list>

namespace osl
{
  namespace search
  {
    typedef std::set<HashKey> KeySet;

    /**
     * TableAnalyzer private namespace.
     * Ū̾Υ饹Τ osl ϱ
     * ǽʤä顤ĤǤʬΥƥƥȤǤ褦
     * ܤ˵ǽ̤˥饹
     */ 
    namespace TA
    {
      MoveLogProb getBestMove(const SimpleHashTable& table, HashEffectState& state)
      {
	const SimpleHashRecord *record = table.find(state.getHash());

	if (record)
	  return record->bestMove;
	return MoveLogProb();
      }

      struct MoveList : public std::list<MoveLogProb>
      {
      };

      void getBestMoveListAfterMove(const SimpleHashTable& table, 
				    HashEffectState& state, 
				    MoveList& moves, Move m, KeySet& dejavu)
      {
	DoUndoMoveStack<HashEffectState> locks;
	while (true)
	{
	  assert(state.isValidMove(m, true));
	  locks.push(state, m);

	  const MoveLogProb bestMove = getBestMove(table, state);
	  if ((! bestMove.validMove())
	      || (dejavu.find(state.getHash()) != dejavu.end()))
	    return;
	  moves.push_back(bestMove); // ̵ͤξ MOVE_INVALID Ͽ
	  if (! bestMove.getMove())
	    return;
	  dejavu.insert(state.getHash());
	  m = bestMove.getMove();
	}
      }

      const int INVALID_VALUE = -1; // FIXME:  Ȥꤢ
      struct MoveEntry 
      {
	Move bestMove;
	Move nextBestMove;
	int upperBound, lowerBound;
	int upperLimit, lowerLimit;
	MoveEntry(Move b, const SimpleHashRecord *record) 
	  : bestMove(b), upperBound(INVALID_VALUE), lowerLimit(INVALID_VALUE)
	{
	  if (record)
	  {
	    nextBestMove = record->bestMove.getMove();
	    upperLimit = record->upperLimit;
	    if (upperLimit > 0)
	      upperBound = record->upperBound;
	    lowerLimit = record->lowerLimit;
	    if (lowerLimit > 0)
	      lowerBound = record->lowerBound;
	  }
	}
      };
      template <Player P>
      struct MoveMap 
	: public std::multimap<int,MoveEntry,EvalCompareLarger<P> > 
      {
      };

      template <Player P>
      struct TopLevelAnalyzer
      {
	const SimpleHashTable *table;
	typedef MoveMap<P> map_t;
	MoveVector unexaminedMoves;
	map_t sorter;
	int numBestMoves;
	Move curMove;
	std::ostream& os;

	TopLevelAnalyzer(const SimpleHashTable *t, int num, std::ostream& o) 
	  : table(t), numBestMoves(num), os(o)
	{
	}
	void examineAll(HashEffectState& state, const MoveVector& moves, const Move bestMove);
      };

    } // namespace TA
  } // namespace search
} // namespace osl

template <osl::Player P>
void osl::search::TA::TopLevelAnalyzer<P>::
examineAll(HashEffectState& state, const MoveVector& moves,
	   const Move bestMove)
{
  for (size_t i=0; i<moves.size(); ++i)
  {
    const Move curMove = moves[i];

    /** ɽϿƤɤȽ˻ȤäƤ */
    int value=INVALID_VALUE;
    /** ɽˤʤĹˡ */
    bool dangerouslyUnexamined=false; 

    DoUndoMoveLock lock(state, curMove);
    const SimpleHashRecord *record = table->find(state.getHash());
    if (curMove == bestMove)
    {
      if (record && record->hasUpperBound(1))
	value = record->upperBound+EvalTraits<P>::delta*2;
      else
	dangerouslyUnexamined = true;
    }
    else
    {
      // P ʤΤ LowerBound  P ˤȤäƤ UpperBound
      if (record && record->hasLowerBound(1))
	value = record->lowerBound;
      else 
      {
	// error 
	const Position myKingPosition
	  = state.getKingPosition(alt(state.getTurn()));
	if (! state.hasEffectBy(state.getTurn(), myKingPosition))
	  dangerouslyUnexamined = true;
      }
    }
    if (value != INVALID_VALUE)
      sorter.insert(std::make_pair
		    (value, MoveEntry(curMove, record)));
    if (dangerouslyUnexamined)
      unexaminedMoves.push_back(curMove);
  }

  if (bestMove && (sorter.begin()->second.bestMove != bestMove))
  {
    os << "possible search bug\n";
    os << sorter.begin()->second.bestMove 
       << " != " << bestMove << "\n";
    os << state;
  }
      
  int numPrinted = 0;
  for (typename map_t::const_iterator p=sorter.begin(); 
       (p!=sorter.end())&&(numPrinted < numBestMoves); ++p, ++numPrinted)
  {
    os 
#if 0
      << std::setw(5) << p->first << " "
#endif
      << "[" 
      << std::setw(5) << p->second.lowerBound << "(" 
      << std::setw(4) << p->second.lowerLimit << ") < " 
      << std::setw(5) << p->second.upperBound << "("
      << std::setw(4) << p->second.upperLimit << ") " 
      << "] \t";
    csaShow(os, p->second.bestMove);
    os << " ";
    // csaShow(os, p->second.nextBestMove);
    MoveList bestMoves;
    KeySet dejavu;
    getBestMoveListAfterMove(*table, state, bestMoves, p->second.bestMove,
			     dejavu);
    for (MoveList::const_iterator p=bestMoves.begin(); p!=bestMoves.end(); ++p)
    {
      if (p->getMove())
      {
	csaShow(os, p->getMove());
	os << " "; // << p->getLogProb() << " ";
      }
      else
	os << "CHECKMATE";
    }
    os << "\n";
  }

  if (unexaminedMoves.size())
  {
    os << "possible move generation failure ";
    for (size_t i=0; i<unexaminedMoves.size(); ++i)
    {
      csaShow(os, unexaminedMoves[i]);
      os << " ";
    }
    os << "\n";
  }
}
  
osl::search::
TableAnalyzer::TableAnalyzer(const SimpleHashTable& t, std::ostream& o) 
  : table(t), os(o)
{
}
osl::search::
TableAnalyzer::~TableAnalyzer()
{
}
  
void osl::search::
TableAnalyzer::examineTopLevel(HashEffectState& state, int numBestMoves) const
{
  os << "analysis of root node\n";
  const MoveLogProb bestMove = TA::getBestMove(table, state);
  MoveVector moves;
  {
    move_action::Store store(moves);
    move_generator::AllMoves<HashEffectState::simple_board_t, move_action::Store>::
      generateMoves(state.getTurn(),state,store);
  }
  if (state.getTurn() == BLACK)
  {
    TA::TopLevelAnalyzer<BLACK> helper(&table, numBestMoves, os);
    helper.examineAll(state, moves, bestMove.getMove());
  }
  else
  {
    TA::TopLevelAnalyzer<WHITE> helper(&table, numBestMoves, os);
    helper.examineAll(state, moves, bestMove.getMove());
  }
}
void osl::search::
TableAnalyzer::examineUpperBounds(HashEffectState& state, int numBestMoves) const
{
  os << "analysis of a node after the best move\n";
  MoveVector moves;
  {
    move_action::Store store(moves);
    move_generator::AllMoves<HashEffectState::simple_board_t, move_action::Store>::
      generateMoves(state.getTurn(),state,store);
  }
  const Move bestMove = MOVE_INVALID;
  if (state.getTurn() == BLACK)
  {
    TA::TopLevelAnalyzer<BLACK> helper(&table, numBestMoves, os);
    helper.examineAll(state, moves, bestMove);
  }
  else
  {
    TA::TopLevelAnalyzer<WHITE> helper(&table, numBestMoves, os);
    helper.examineAll(state, moves, bestMove);
  }
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
